@@ -1,14 +1,17 @@
-from flask.ext.restful import Resource, fields
-from flask import request, abort, Response
-import inspect
 import functools
-import re
-from flask_restful_swagger import registry, registered, api_spec_endpoint
-from jinja2 import Template
+import inspect
 import os
+import re
+
+from flask import request, abort, Response
+from flask.ext.restful import Resource, fields
+from flask_restful_swagger import (
+  registry, api_spec_static)
+from jinja2 import Template
 
 resource_listing_endpoint = None
 
+
 def docs(api, apiVersion='0.0', swaggerVersion='1.2',
          basePath='http://localhost:5000',
          resourcePath='/',
@@ -18,18 +21,23 @@ def docs(api, apiVersion='0.0', swaggerVersion='1.2',
   api_add_resource = api.add_resource
 
   def add_resource(resource, path, *args, **kvargs):
-    endpoint = swagger_endpoint(resource, path)
+    register_once(api, api_add_resource, apiVersion, swaggerVersion, basePath,
+                  resourcePath, produces, api_spec_url)
+
+    resource = make_class(resource)
+    endpoint = swagger_endpoint(api, resource, path)
+
     # Add a .help.json help url
     swagger_path = extract_swagger_path(path)
-    endpoint_path = "%s_help_json" % resource.__name__
-    api_add_resource(endpoint, "%s.help.json" % swagger_path,
-                     endpoint=endpoint_path)
+
     # Add a .help.html help url
-    endpoint_path = "%s_help_html" % resource.__name__
-    api_add_resource(endpoint, "%s.help.html" % swagger_path,
-                     endpoint=endpoint_path)
-    register_once(api_add_resource, apiVersion, swaggerVersion, basePath,
-                  resourcePath, produces, api_spec_url)
+    endpoint_html_str = '{0}/help'.format(swagger_path)
+    api_add_resource(
+      endpoint,
+      "{0}.help.json".format(swagger_path),
+      "{0}.help.html".format(swagger_path),
+      endpoint=endpoint_html_str)
+
     return api_add_resource(resource, path, *args, **kvargs)
 
   api.add_resource = add_resource
@@ -38,69 +46,161 @@ def docs(api, apiVersion='0.0', swaggerVersion='1.2',
 
 rootPath = os.path.dirname(__file__)
 
-def register_once(add_resource_func, apiVersion, swaggerVersion, basePath,
-                  resourcePath, produces, endpoint):
-  global registered
-  global api_spec_endpoint
+
+def make_class(class_or_instance):
+  if inspect.isclass(class_or_instance):
+    return class_or_instance
+  return class_or_instance.__class__
+
+
+def register_once(api, add_resource_func, apiVersion, swaggerVersion, basePath,
+                  resourcePath, produces, endpoint_path):
   global api_spec_static
   global resource_listing_endpoint
-  if not registered:
-    registered = True
-    registry['apiVersion'] = apiVersion
-    registry['swaggerVersion'] = swaggerVersion
-    registry['basePath'] = basePath
-    registry['resourcePath'] = resourcePath
-    registry['produces'] = produces
-    add_resource_func(SwaggerRegistry, endpoint, endpoint=endpoint)
-    api_spec_endpoint = endpoint + '.json'
-    add_resource_func(SwaggerRegistry, api_spec_endpoint, endpoint=api_spec_endpoint)
-    ep = endpoint + '.html'
-    add_resource_func(SwaggerRegistry, ep, endpoint=ep)
-    resource_listing_endpoint = endpoint + '/_/resource_list.json'
-    add_resource_func(ResourceLister, resource_listing_endpoint, endpoint=resource_listing_endpoint)
-    api_spec_static = endpoint + '/_/static/'
-    add_resource_func(StaticFiles, api_spec_static + '<string:dir1>/<string:dir2>/<string:dir3>', api_spec_static + '<string:dir1>/<string:dir2>', api_spec_static + '<string:dir1>')
+
+  if api.blueprint and not registry.get(api.blueprint.name):
+    # Most of all this can be taken from the blueprint/app
+    registry[api.blueprint.name] = {
+      'apiVersion': apiVersion,
+      'swaggerVersion': swaggerVersion,
+      'basePath': basePath,
+      'spec_endpoint_path': endpoint_path,
+      'resourcePath': resourcePath,
+      'produces': produces,
+      'x-api-prefix': '',
+      'apis': []
+    }
+
+    def registering_blueprint(setup_state):
+      reg = registry[setup_state.blueprint.name]
+      reg['x-api-prefix'] = setup_state.url_prefix
+
+    api.blueprint.record(registering_blueprint)
+
+    add_resource_func(
+      SwaggerRegistry,
+      endpoint_path,
+      endpoint_path + '.json',
+      endpoint_path + '.html'
+    )
+
+    resource_listing_endpoint = endpoint_path + '/_/resource_list.json'
+    add_resource_func(ResourceLister, resource_listing_endpoint)
+
+    api_spec_static = endpoint_path + '/_/static/'
+    add_resource_func(
+      StaticFiles,
+      api_spec_static + '<string:dir1>/<string:dir2>/<string:dir3>',
+      api_spec_static + '<string:dir1>/<string:dir2>',
+      api_spec_static + '<string:dir1>')
+  elif not 'app' in registry:
+    registry['app'] = {
+      'apiVersion': apiVersion,
+      'swaggerVersion': swaggerVersion,
+      'basePath': basePath,
+      'spec_endpoint_path': endpoint_path,
+      'resourcePath': resourcePath,
+      'produces': produces,
+    }
+
+    add_resource_func(
+      SwaggerRegistry,
+      endpoint_path,
+      endpoint_path + '.json',
+      endpoint_path + '.html',
+      endpoint='app/registry'
+    )
+
+    resource_listing_endpoint = endpoint_path + '/_/resource_list.json'
+    add_resource_func(
+      ResourceLister, resource_listing_endpoint,
+      endpoint='app/resourcelister')
+
+    api_spec_static = endpoint_path + '/_/static/'
+    add_resource_func(
+      StaticFiles,
+      api_spec_static + '<string:dir1>/<string:dir2>/<string:dir3>',
+      api_spec_static + '<string:dir1>/<string:dir2>',
+      api_spec_static + '<string:dir1>',
+      endpoint='app/staticfiles')
 
 templates = {}
 
+
 def render_endpoint(endpoint):
   return render_page("endpoint.html", endpoint.__dict__)
 
+
 def render_homepage(resource_list_url):
   conf = {'resource_list_url': resource_list_url}
   return render_page("index.html", conf)
 
+
+def _get_current_registry(api=None):
+  # import ipdb;ipdb.set_trace()
+  global registry
+  app_name = None
+  overrides = {}
+  if api:
+    app_name = api.blueprint.name if api.blueprint else None
+  else:
+    app_name = request.blueprint
+    overrides = {'basePath': request.url_root.rstrip('/')}
+
+  if not app_name:
+    app_name = 'app'
+
+  overrides['models'] = registry.get('models', {})
+
+  reg = registry.setdefault(app_name, {})
+  reg.update(overrides)
+
+  reg['basePath'] = reg['basePath'] + reg.get('x-api-prefix', '')
+
+  return reg
+
+
 def render_page(page, info):
-  url = registry['basePath']
+  req_registry = _get_current_registry()
+  url = req_registry['basePath']
   if url.endswith('/'):
-    url = url.rtrim('/')
-  conf = {'base_url': api_spec_static, 'full_base_url': url + api_spec_static}
-  if info != None:
+    url = url.rstrip('/')
+  conf = {
+    'base_url': url + api_spec_static,
+    'full_base_url': url + api_spec_static
+  }
+  if info is not None:
     conf.update(info)
   global templates
-  if templates.has_key(page):
+  if page in templates:
     template = templates[page]
   else:
-    fs = open(os.path.join(rootPath, 'static', page), "r")
-    template = Template(fs.read())
-    templates[page] = template
+    with open(os.path.join(rootPath, 'static', page), "r") as fs:
+      template = Template(fs.read())
+      templates[page] = template
   mime = 'text/html'
   if page.endswith('.js'):
     mime = 'text/javascript'
   return Response(template.render(conf), mimetype=mime)
 
+
 class StaticFiles(Resource):
-  def get(self, dir1 = None, dir2 = None, dir3 = None):
-    if dir1 == None:
+
+  def get(self, dir1=None, dir2=None, dir3=None):
+    req_registry = _get_current_registry()
+
+    if dir1 is None:
       filePath = "index.html"
     else:
       filePath = dir1
-      if dir2 != None:
+      if dir2 is not None:
         filePath = "%s/%s" % (filePath, dir2)
-        if dir3 != None:
+        if dir3 is not None:
           filePath = "%s/%s" % (filePath, dir3)
-    if filePath in ["index.html", "o2c.html", "swagger-ui.js", "swagger-ui.min", "lib/swagger-oauth.js"]:
-      conf = {'resource_list_url': api_spec_endpoint}
+    if filePath in [
+      "index.html", "o2c.html", "swagger-ui.js"
+       "swagger-ui.min", "lib/swagger-oauth.js"]:
+      conf = {'resource_list_url': req_registry['spec_endpoint_path']}
       return render_page(filePath, conf)
     mime = 'text/plain'
     if filePath.endswith(".gif"):
@@ -117,23 +217,27 @@ class StaticFiles(Resource):
       return Response(fs, mimetype=mime)
     abort(404)
 
+
 class ResourceLister(Resource):
   def get(self):
+    req_registry = _get_current_registry()
     return {
-      "apiVersion": registry['apiVersion'],
-      "swaggerVersion": registry['swaggerVersion'],
+      "apiVersion": req_registry['apiVersion'],
+      "swaggerVersion": req_registry['swaggerVersion'],
       "apis": [
         {
-          "path": '/..' * (len(api_spec_endpoint.split('/')) + 1)  + api_spec_endpoint,
+          "path": (
+            req_registry['basePath'] + req_registry['spec_endpoint_path']),
           "description": "Auto generated API docs by flask-restful-swagger"
         }
       ]
     }
 
 
-def swagger_endpoint(resource, path):
+def swagger_endpoint(api, resource, path):
   endpoint = SwaggerEndpoint(resource, path)
-  registry['apis'].append(endpoint.__dict__)
+  req_registry = _get_current_registry(api=api)
+  req_registry.setdefault('apis', []).append(endpoint.__dict__)
 
   class SwaggerResource(Resource):
     def get(self):
@@ -144,11 +248,30 @@ def swagger_endpoint(resource, path):
   return SwaggerResource
 
 
+def _sanitize_doc(comment):
+  return comment.replace('\n', '<br/>') if comment else comment
+
+
+def _parse_doc(obj):
+  first_line, other_lines = None, None
+
+  full_doc = inspect.getdoc(obj)
+  if full_doc:
+    line_feed = full_doc.find('\n')
+    if line_feed != -1:
+      first_line = _sanitize_doc(full_doc[:line_feed])
+      other_lines = _sanitize_doc(full_doc[line_feed+1:])
+    else:
+      first_line = full_doc
+
+  return first_line, other_lines
+
+
 class SwaggerEndpoint(object):
   def __init__(self, resource, path):
     self.path = extract_swagger_path(path)
     path_arguments = extract_path_arguments(path)
-    self.description = inspect.getdoc(resource)
+    self.description, self.notes = _parse_doc(resource)
     self.operations = self.extract_operations(resource, path_arguments)
 
   @staticmethod
@@ -166,21 +289,26 @@ class SwaggerEndpoint(object):
           'parameters': path_arguments,
           'nickname': 'nickname'
       }
-      op['summary'] = inspect.getdoc(method_impl)
+      op['summary'], op['notes'] = _parse_doc(method_impl)
+
       if '__swagger_attr' in method_impl.__dict__:
         # This method was annotated with @swagger.operation
         decorators = method_impl.__dict__['__swagger_attr']
         for att_name, att_value in list(decorators.items()):
           if isinstance(att_value, (str, int, list)):
             if att_name == 'parameters':
-              op['parameters'] = merge_parameter_list(op['parameters'], att_value)
+              op['parameters'] = merge_parameter_list(
+                op['parameters'], att_value)
             else:
+              if att_name in op:
+                att_value = '{0}<br/>{1}'.format(att_value, op[att_name])
               op[att_name] = att_value
           elif isinstance(att_value, object):
             op[att_name] = att_value.__name__
       operations.append(op)
     return operations
 
+
 def merge_parameter_list(base, override):
   base = list(base)
   names = [x['name'] for x in base]
@@ -193,11 +321,14 @@ def merge_parameter_list(base, override):
       base.append(o)
   return base
 
+
 class SwaggerRegistry(Resource):
   def get(self):
+    req_registry = _get_current_registry()
     if request.path.endswith('.html'):
-      return render_homepage(resource_listing_endpoint)
-    return registry
+      return render_homepage(
+        req_registry['basePath'] + req_registry['spec_endpoint_path'] + '/_/resource_list.json')
+    return req_registry
 
 
 def operation(**kwargs):
@@ -217,6 +348,7 @@ def model(c=None, *args, **kwargs):
   add_model(c)
   return c
 
+
 class _Nested(object):
   def __init__(self, klass, **kwargs):
     self._nested = kwargs
@@ -228,6 +360,7 @@ class _Nested(object):
   def nested(self):
     return self._nested
 
+
 # wrap _Cache to allow for deferred calling
 def nested(klass=None, **kwargs):
   if klass:
@@ -241,11 +374,12 @@ def nested(klass=None, **kwargs):
     ret = wrapper
   return ret
 
+
 def add_model(model_class):
   models = registry['models']
   name = model_class.__name__
   model = models[name] = {'id': name}
-  model['description'] = inspect.getdoc(model_class)
+  model['description'], model['notes'] = _parse_doc(model_class)
   if 'resource_fields' in dir(model_class):
     # We take special care when the model class has a field resource_fields.
     # By convension this field specifies what flask-restful would return when
@@ -274,7 +408,8 @@ def add_model(model_class):
     defaults = {}
     required = model['required'] = []
     if argspec.defaults:
-      defaults = list(zip(argspec.args[-len(argspec.defaults):], argspec.defaults))
+      defaults = list(
+        zip(argspec.args[-len(argspec.defaults):], argspec.defaults))
     properties = model['properties'] = {}
     for arg in argspec.args[:-len(defaults)]:
       required.append(arg)
@@ -311,12 +446,14 @@ def deduce_swagger_type(python_type_or_object, nested_type=None):
         else:
           return {'type': 'array',
                   'items': {
-                    '$ref': deduce_swagger_type_flat(python_type_or_object.container, nested_type)}}
+                    '$ref': deduce_swagger_type_flat(
+                      python_type_or_object.container, nested_type)}}
     if predicate(python_type_or_object, (fields.Nested)):
         return {'type': nested_type}
 
     return {'type': 'null'}
 
+
 def deduce_swagger_type_flat(python_type_or_object, nested_type=None):
     if nested_type:
       return nested_type
@@ -374,6 +511,7 @@ def extract_path_arguments(path):
     spl = arg.split(':')
     if len(spl) == 1:
       return {'name': spl[0],
+              'dataType': 'string',
               'paramType': 'path'}
     else:
       return {'name': spl[1],
